"""
Graphviz tools tests

The test cases in this file are sanity checks on the various tools in the
Graphviz suite. A failure of one of these indicates that some basic functional
property of one of the tools has been broken.
"""

import os
import platform
import re
import subprocess
import sys
import xml.etree.ElementTree as ET
from pathlib import Path

import pytest

sys.path.append(os.path.dirname(__file__))
from gvtest import (  # pylint: disable=wrong-import-position
    has_sandbox,
    remove_asan_summary,
    remove_xtype_warnings,
    run,
    run_raw,
    which,
)


@pytest.mark.parametrize(
    "tool",
    [
        "acyclic",
        "bcomps",
        "ccomps",
        "circo",
        "cluster",
        "dijkstra",
        "dot",
        "dot2gxl",
        "dot_builtins",
        "edgepaint",
        "fdp",
        "gc",
        "gml2gv",
        "graphml2gv",
        "gv2gml",
        "gv2gxl",
        "gvcolor",
        "gvedit",
        "gvgen",
        "gvmap",
        "gvmap.sh",
        "gvpack",
        "gvpr",
        "gxl2dot",
        "gxl2gv",
        "mingle",
        "mm2gv",
        "neato",
        "nop",
        "osage",
        "patchwork",
        "prune",
        "sccmap",
        "sfdp",
        "smyrna",
        "tred",
        "twopi",
        "unflatten",
        "vimdot",
    ],
)
def test_tools(tool):
    """
    check the help/usage output of the given tool looks correct
    """

    if which(tool) is None:
        pytest.skip(f"{tool} not available")

    # exec-ing a POSIX shell script as-is does not work on Windows
    if tool == "gvmap.sh" and platform.system() == "Windows":
        pytest.skip("gvmap.sh cannot be run directly on Windows")

    # Ensure that X fails to open display
    environ_copy = os.environ.copy()
    environ_copy.pop("DISPLAY", None)

    # Test usage
    with subprocess.Popen(
        [tool, "-?"],
        env=environ_copy,
        stdin=subprocess.DEVNULL,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        text=True,
    ) as p:
        output, _ = p.communicate()
        ret = p.returncode

    assert ret == 0, f"`{tool} -?` failed. Output was: {output}"

    output = remove_asan_summary(output)
    output = remove_xtype_warnings(output)
    assert (
        re.match("usage", output, flags=re.IGNORECASE) is not None
    ), f"{tool} -? did not show usage. Output was: {output}"

    # Test unsupported option
    returncode = subprocess.call(
        [tool, "-$"],
        env=environ_copy,
    )

    assert returncode != 0, f"{tool} accepted unsupported option -$"


@pytest.mark.skipif(which("edgepaint") is None, reason="edgepaint not available")
@pytest.mark.parametrize(
    "arg",
    (
        "-accuracy=0.01",
        "-angle=15",
        "-random_seed=42",
        "-random_seed=-42",
        "-lightness=0,70",
        "-share_endpoint",
        f"-o {os.devnull}",
        "-color_scheme=accent7",
        "-v",
    ),
)
def test_edgepaint_options(arg: str):
    """
    edgepaint should correctly understand all its command line flags
    """

    # a basic graph that edgepaint can process
    input = (
        "digraph {\n"
        '  graph [bb="0,0,54,108"];\n'
        '  node [label="\\N"];\n'
        "  a       [height=0.5,\n"
        '           pos="27,90",\n'
        "           width=0.75];\n"
        "  b       [height=0.5,\n"
        '           pos="27,18",\n'
        "           width=0.75];\n"
        '  a -> b  [pos="e,27,36.104 27,71.697 27,63.983 27,54.712 27,46.112"];\n'
        "}"
    )

    # run edgepaint on this
    args = ["edgepaint"] + arg.split(" ")
    try:
        run(*args, input=input)
    except subprocess.CalledProcessError as e:
        raise RuntimeError(f"edgepaint rejected command line option '{arg}'") from e


@pytest.mark.skipif(not has_sandbox(), reason="no supported sandbox available")
def test_sandbox_noop():
    """check trivial functionality works when sandboxed"""
    sandbox = which("dot_sandbox")
    run_raw(sandbox, "-V")


@pytest.mark.skipif(not has_sandbox(), reason="no supported sandbox available")
def test_sandbox_basic():
    """check processing a simple graph when sandboxed"""
    sandbox = which("dot_sandbox")
    run(sandbox, input="graph { a -- b; }")


@pytest.mark.skipif(not has_sandbox(), reason="no supported sandbox available")
def test_sandbox_render():
    """check rendering works when sandboxed"""
    sandbox = which("dot_sandbox")
    stdout = run(sandbox, "-Tsvg", input="graph { a -- b; }")
    ET.fromstring(stdout)


@pytest.mark.skipif(not has_sandbox(), reason="no supported sandbox available")
def test_sandbox_write(tmp_path: Path):
    """check file writing is prevented when sandboxed"""
    sandbox = which("dot_sandbox")
    proc = subprocess.run(
        [sandbox, "-o", "probe.dot"],
        input="graph { a -- b; }",
        text=True,
        cwd=tmp_path,
        check=False,
    )
    assert not (
        tmp_path / "probe.dot"
    ).exists(), "file writing within sandbox was not prevented"
    assert (
        proc.returncode != 0
    ), "failed file writing did not cause a non-zero exit status"


@pytest.mark.skipif(not has_sandbox(), reason="no supported sandbox available")
def test_sandbox_write_2(tmp_path: Path):
    """check file writing via `-O` is prevented when sandboxed"""
    sandbox = which("dot_sandbox")
    proc = subprocess.run(
        [sandbox, "-Tsvg", "-O"],
        input="graph { a -- b; }",
        text=True,
        cwd=tmp_path,
        check=False,
    )
    assert not (
        tmp_path / "noname.gv.svg"
    ).exists(), "file writing within sandbox was not prevented"
    assert (
        proc.returncode != 0
    ), "failed file writing did not cause a non-zero exit status"



#!/usr/bin/env python3

"""
convert a json file to an html file
"""

import argparse
import json

from json2html import json2html


def main():
    """entry point"""

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("input", type=argparse.FileType("rt"), help="Input file")
    parser.add_argument("output", type=argparse.FileType("wt"), help="Output file")
    args = parser.parse_args()

    json_input = json.load(args.input)
    html_output = json2html.convert(
        json=json_input, table_attributes='class="jsontable"'
    )
    args.output.write(html_output)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""
test cases that are only relevant to run in CI
"""

import os
import platform
import subprocess
import sys

import pytest

sys.path.append(os.path.join(os.path.dirname(__file__), "../tests"))
from gvtest import (  # pylint: disable=wrong-import-position
    ROOT,
    build_system,
    dot,
    freedesktop_os_release,
    is_cmake,
    is_macos,
    is_mingw,
    is_static_build,
    is_ubuntu_2604,
    run,
    which,
)


@pytest.mark.parametrize(
    "binary",
    [
        "acyclic",
        "bcomps",
        "ccomps",
        "circo",
        "cluster",
        "dijkstra",
        "dot",
        "dot2gxl",
        "dot_builtins",
        "edgepaint",
        "fdp",
        "gc",
        "gml2gv",
        "graphml2gv",
        "gv2gml",
        "gv2gxl",
        "gvcolor",
        "gvedit",
        "gvgen",
        "gvmap",
        "gvmap.sh",
        "gvpack",
        "gvpr",
        "gxl2dot",
        "gxl2gv",
        "mingle",
        "mm2gv",
        "neato",
        "nop",
        "osage",
        "patchwork",
        "prune",
        "sccmap",
        "sfdp",
        "smyrna",
        "tred",
        "twopi",
        "unflatten",
        "vimdot",
    ],
)
def test_existence(binary: str):
    """
    check that a given binary was built and is on $PATH
    """

    os_id = freedesktop_os_release().get("ID")

    # https://gitlab.com/graphviz/graphviz/-/issues/1834
    if os_id == "rocky" and binary == "smyrna":
        check_that_tool_does_not_exist(binary, os_id)
        pytest.skip("smyrna is not built for Rocky due to lacking libgts")

    if os_id == "rocky" and binary == "mingle":
        check_that_tool_does_not_exist(binary, os_id)
        pytest.skip("mingle is not built for Rocky due to lacking libANN")

    if binary == "mingle" and is_mingw():
        check_that_tool_does_not_exist(binary, os_id)
        pytest.skip(f"{binary} is not built on some Windows due to lacking libANN")

    if binary == "gvedit" and platform.system() == "Windows" and not is_mingw():
        check_that_tool_does_not_exist(binary, os_id)
        pytest.skip(f"{binary} is not built on Windows due to lacking Qt")

    # Smyrna dependencies are not avaiable in other jobs
    if binary == "smyrna" and platform.system() != "Linux":
        check_that_tool_does_not_exist(binary, os_id)
        pytest.skip("smyrna is not built on non-Linux due to lacking dependencies")

    if binary == "smyrna" and is_ubuntu_2604():
        check_that_tool_does_not_exist(binary, os_id)
        pytest.skip("smyrna is not built on Ubuntu ≥ 25.10 due to lacking dependencies")

    if binary == "gvmap.sh" and platform.system() == "Windows":
        pytest.skip(f"{binary} detection is unreliable on Windows")

    if binary == "vimdot" and platform.system() == "Windows":
        check_that_tool_does_not_exist(binary, os_id)
        pytest.skip(f"{binary} is not installed on Windows")

    if binary == "vimdot" and not is_cmake() and is_macos():
        check_that_tool_does_not_exist(binary, os_id)
        pytest.skip("https://gitlab.com/graphviz/graphviz/-/issues/2423")

    if binary == "dot_builtins" and is_static_build():
        pytest.skip("dot_builtins may not be built in a static build")

    if binary == "diffimg" and build_system() == "cmake" and is_mingw():
        pytest.skip("libgd not detected on CMake+MinGW")

    assert which(binary) is not None


def check_that_tool_does_not_exist(tool, os_id):
    """
    validate that the given tool does *not* exist
    """
    assert which(tool) is None, (
        f"{tool} has been resurrected in the "
        f"{build_system()} build on {os_id}. Please remove skip."
    )


@pytest.mark.skipif(
    build_system() == "cmake" and is_mingw(), reason="libgd not detected on CMake+MinGW"
)
def test_1786():
    """
    png:gd format should be supported
    https://gitlab.com/graphviz/graphviz/-/issues/1786
    """

    # run a trivial graph through Graphviz
    dot("png:gd", source="digraph { a -> b; }")


def test_2469():
    """
    svgz format should be supported
    https://gitlab.com/graphviz/graphviz/-/issues/2469
    """

    # run a trivial graph through Graphviz
    dot("svgz", source="digraph { a -> b; }")


def test_installation():
    """
    check that Graphviz reports the expected version number
    """

    expected_version = run(sys.executable, ROOT / "gen_version.py").strip()

    actual_version_string = run("dot", "-V", stderr=subprocess.STDOUT)
    try:
        actual_version = actual_version_string.split()[4]
    except IndexError:
        pytest.fail(f"Malformed version string: {actual_version_string}")
    assert actual_version == expected_version



"""test ../lib/cgraph/ internal generic utility code"""

import os
import platform
import subprocess
import sys
from pathlib import Path

import pytest

sys.path.append(os.path.dirname(__file__))
from gvtest import (  # pylint: disable=wrong-import-position
    compile_c,
    is_mingw,
    run,
    run_c,
)


@pytest.mark.parametrize("utility", ("arena", "bitarray", "itos", "list", "tokenize"))
def test_utility(utility: str, tmp_path: Path):
    """run the given utility’s unit tests"""

    # locate the unit tests
    src = Path(__file__).parent.resolve() / f"../lib/util/test_{utility}.c"
    assert src.exists()

    # locate lib directory that needs to be in the include path
    lib = Path(__file__).parent.resolve() / "../lib"

    # extra C flags this compilation needs
    cflags = ["-I", lib]
    if platform.system() != "Windows":
        cflags += ["-std=gnu17"]

    _, _ = run_c(src, tmp_path, cflags=cflags)


@pytest.mark.parametrize("builtins", (False, True))
def test_overflow_h(builtins: bool, tmp_path: Path):
    """test ../lib/util/overflow.h"""

    # locate the unit test
    src = Path(__file__).parent.resolve() / "../lib/util/test_overflow.c"
    assert src.exists()

    # locate lib directory that needs to be in the include path
    lib = Path(__file__).parent.resolve() / "../lib"

    # extra C flags this compilation needs
    cflags = ["-I", lib]
    if not builtins:
        cflags += ["-DSUPPRESS_BUILTINS"]
    if platform.system() != "Windows":
        cflags += ["-std=gnu17"]

    run_c(src, tmp_path, cflags=cflags)


@pytest.mark.parametrize(
    "bad_test",
    (
        1,
        2,
        3,
        4,
        pytest.param(
            5,
            marks=pytest.mark.skipif(
                platform.system() == "Windows" and not is_mingw(),
                reason="this test has not been ported to Windows",
            ),
        ),
        pytest.param(
            6,
            marks=pytest.mark.skipif(
                platform.system() == "Windows" and not is_mingw(),
                reason="this test has not been ported to Windows",
            ),
        ),
    ),
)
def test_list_type_safetyutility(bad_test: int):
    """
    check the type safety of ../lib/util/list.h’s interfaces

    Args:
        bad_test: Which type-violating test in ../lib/util/test_list.c to compile.
    """

    # locate the unit tests
    src = Path(__file__).parent.resolve() / "../lib/util/test_list.c"
    assert src.exists()

    # locate lib directory that needs to be in the include path
    lib = Path(__file__).parent.resolve() / "../lib"

    # extra C flags this compilation needs
    cflags = [f"-DBAD_TEST={bad_test}", "-I", lib]
    if platform.system() != "Windows":
        cflags += ["-std=gnu17"]

    # this should fail to compile if strong type safety is preserved
    with pytest.raises(subprocess.CalledProcessError):
        compile_c(src, cflags=cflags)


def test_gv_find_me(tmp_path: Path):
    """test ../lib/util/gv_find_me.c::gv_find_me works correctly"""

    # locate our support test file
    src = Path(__file__).parent.resolve() / "test_gv_find_me.c"
    assert src.exists()

    # locate lib directory that needs to be in the include path
    lib = Path(__file__).resolve().parents[1] / "lib"

    # extra C flags this compilation needs
    cflags = ["-I", lib]
    if platform.system() != "Windows":
        cflags += ["-std=gnu17"]

    # pick a path to compile to
    exe = tmp_path / "test_gv_find_me.exe"

    _ = compile_c(src, cflags, dst=exe)

    # run this
    output = run(exe)

    assert output == f"{exe}\n", "gv_find_me did not determine executable absolute path"



#!/usr/bin/env python3

"""
example usage of the Graphviz Python module
"""

import gv

# create a new empty graph
G = gv.digraph("G")

# define a simple graph ( A->B )
gv.edge(gv.node(G, "A"), gv.node(G, "B"))

# compute a directed graph layout
gv.layout(G, "dot")

# annotate the graph with the layout information
gv.render(G, "ps")



#!/usr/bin/env python3

"""
example usage of the Graphviz Python module
"""

import gv


def main():
    """entry point"""

    # create a new empty graph
    G = gv.digraph("G")
    # define a simple graph ( A->B )
    gv.edge(gv.node(G, "A"), gv.node(G, "B"))
    # compute a directed graph layout
    gv.layout(G, "dot")
    # annotate the graph with the layout information
    gv.render(G)
    # do something with the layout
    n = gv.firstnode(G)
    while n:
        print(f"node {gv.nameof(n)} is at {gv.getv(n, 'pos')}")
        e = gv.firstout(n)
        while e:
            print(
                f"edge {gv.nameof(gv.tailof(e))}->{gv.nameof(gv.headof(e))} "
                f"is at {gv.getv(e, 'pos')}"
            )
            e = gv.nextout(n, e)
        n = gv.nextnode(G, n)


if __name__ == "__main__":
    main()



"""tests that examples provided with Graphviz compile correctly"""

import os
import platform
import subprocess
import sys
from pathlib import Path

import pytest

sys.path.append(os.path.dirname(__file__))
from gvtest import (  # pylint: disable=wrong-import-position
    is_static_build,
    run,
    run_c,
    run_raw,
    which,
)


@pytest.mark.parametrize(
    "src", ["demo.c", "dot.c", "example.c", "neatopack.c", "simple.c"]
)
@pytest.mark.skipif(
    is_static_build(),
    reason="dynamic libraries are unavailable to link against in static builds",
)
def test_compile_example(src, tmp_path):
    """try to compile the example"""

    # construct an absolute path to the example
    filepath = Path(__file__).parent.resolve() / ".." / "dot.demo" / src

    libs = ["cgraph", "gvc"]

    # run the example
    args = ["-Kneato"] if src in ["demo.c", "dot.c"] else []

    if platform.system() == "Windows":
        cflags = ["-DGVDLL"]
    else:
        cflags = None

    _, _ = run_c(filepath, tmp_path, args, "graph {a -- b}", cflags=cflags, link=libs)


@pytest.mark.parametrize(
    "src",
    [
        "addrings",
        "attr",
        "bbox",
        "bipart",
        "chkedges",
        "clustg",
        "collapse",
        "cycle",
        "deghist",
        "delmulti",
        "depath",
        "flatten",
        "group",
        "indent",
        "path",
        "scale",
        "span",
        "treetoclust",
        "addranks",
        "anon",
        "bb",
        "chkclusters",
        "cliptree",
        "col",
        "color",
        "dechain",
        "deledges",
        "delnodes",
        "dijkstra",
        "get-layers-list",
        "knbhd",
        "maxdeg",
        "rotate",
        "scalexy",
        "topon",
    ],
)
@pytest.mark.skipif(which("gvpr") is None, reason="GVPR not available")
def test_gvpr_example(src):
    """check GVPR can parse the given example"""

    # construct a relative path to the example because gvpr on Windows does not
    # support absolute paths (#1780)
    path = Path("cmd/gvpr/lib") / src
    wd = Path(__file__).parent.parent.resolve()

    # run GVPR with the given script
    run_raw("gvpr", "-f", path, stdin=subprocess.DEVNULL, cwd=wd)


@pytest.mark.skipif(which("gvpr") is None, reason="GVPR not available")
def test_gvpr_clustg():
    """check cmd/gvpr/lib/clustg works"""

    # construct a relative path to clustg because gvpr on Windows does not
    # support absolute paths (#1780)
    path = Path("cmd/gvpr/lib/clustg")
    wd = Path(__file__).parent.parent.resolve()

    # some sample input
    input = "digraph { N1; N2; N1 -> N2; N3; }"

    # run GVPR on this input
    output = run("gvpr", "-f", path, input=input, cwd=wd)

    assert (
        output.strip() == 'strict digraph "clust%1" {\n'
        "\tnode [_cnt=0];\n"
        "\tedge [_cnt=0];\n"
        "\tN1 -> N2\t[_cnt=1];\n"
        "\tN3;\n"
        "}"
    ), "unexpected output"



#!/usr/bin/env python3

"""Graphviz test coverage analysis script"""

import argparse
import logging
import shlex
import subprocess
import sys
from pathlib import Path
from typing import Union

# logging output stream, setup in main()
log = None


def run(
    args: list[Union[str, Path]],
) -> None:
    """run a command, echoing it beforehand"""

    print(f"+ {shlex.join(str(x) for x in args)}", flush=True)
    subprocess.check_call(args)


def main(args: list[str]) -> int:
    """entry point"""

    # setup logging to print to stderr
    global log
    ch = logging.StreamHandler()
    log = logging.getLogger("test_coverage.py")
    log.addHandler(ch)

    # parse command line arguments
    parser = argparse.ArgumentParser(
        description="Graphviz test coverage analysis script"
    )
    parser.add_argument(
        "--init",
        action="store_true",
        help="Capture initial zero coverage data before running any test",
    )
    parser.add_argument(
        "--analyze",
        action="store_true",
        help="Analyze test coverage after running tests",
    )
    options = parser.parse_args(args[1:])

    if not options.init and not options.analyze:
        log.error("Must specify --init or --analyze; refusing to run")
        return -1

    cwd = Path.cwd()

    generated_files = [
        cwd / "build/cmd/tools/gmlparse.c",
        cwd / "build/cmd/tools/gmlscan.c",
        cwd / "build/lib/cgraph/grammar.c",
        cwd / "build/lib/cgraph/scan.c",
        cwd / "build/lib/common/htmlparse.c",
        cwd / "build/lib/expr/exparse.c",
        cwd / "build/cmd/gvedit/gvedit_autogen/EWIEGA46WW/moc_csettings.cpp",
        cwd / "build/cmd/gvedit/gvedit_autogen/EWIEGA46WW/moc_imageviewer.cpp",
        cwd / "build/cmd/gvedit/gvedit_autogen/EWIEGA46WW/moc_mainwindow.cpp",
        cwd / "build/cmd/gvedit/gvedit_autogen/EWIEGA46WW/moc_mdichild.cpp",
        cwd / "build/cmd/gvedit/gvedit_autogen/EWIEGA46WW/qrc_mdi.cpp",
    ]

    # files that are generated but only need to be excluded during init
    init_generated_files = [
        cwd / "build/tclpkg/gv/CMakeFiles/gv_d.dir/gvD_wrap.cxx",
        cwd / "build/tclpkg/gv/CMakeFiles/gv_go.dir/gvGO_wrap.cxx",
        cwd / "build/tclpkg/gv/CMakeFiles/gv_sharp.dir/gvCSHARP_wrap.cxx",
        cwd / "build/tclpkg/gv/CMakeFiles/gv_tcl.dir/gvTCL_wrap.cxx",
    ]

    excluded_files = generated_files
    init_excluded_files = init_generated_files

    exclude_options = [f"--exclude={f}" for f in excluded_files]
    init_exclude_options = [f"--exclude={f}" for f in init_excluded_files]

    if options.init:
        run(
            [
                "lcov",
                "--capture",
                "--initial",
                "--directory=.",
                "--branch-coverage",
                "--no-external",
            ]
            + exclude_options
            + init_exclude_options
            + ["--output-file=app_base.info"]
        )

        return 0

    if options.analyze:
        # capture test coverage data
        run(
            [
                "lcov",
                "--capture",
                "--directory=.",
                "--branch-coverage",
                "--no-external",
                "--rc=check_data_consistency=0",
            ]
            + exclude_options
            + ["--output-file=app_test.info"]
        )
        # combine baseline and test coverage data
        run(
            [
                "lcov",
                "--branch-coverage",
                "--add-tracefile=app_base.info",
                "--add-tracefile=app_test.info",
                "--output-file=app_total.info",
                "--rc=check_data_consistency=0",
            ]
        )
        # generate coverage html pages using lcov which are nicer than gcovr's
        Path("coverage/lcov").mkdir(parents=True, exist_ok=True)
        run(
            [
                "genhtml",
                f"--prefix={cwd}",
                "--branch-coverage",
                "--output-directory=coverage/lcov",
                "--rc=check_data_consistency=0",
                "--show-details",
                "app_total.info",
            ]
        )
        # generate coverage info for GitLab's Test Coverage Visualization
        Path("coverage/gcovr").mkdir(parents=True, exist_ok=True)
        run(
            ["gcovr"]
            + exclude_options
            + [f"--gcov-exclude={f}" for f in generated_files]
            + [
                "--xml-pretty",
                "--html-details=coverage/gcovr/index.html",
                "--exclude-unreachable-branches",
                "--gcov-ignore-errors=no_working_dir_found",
                "--print-summary",
                "--output=coverage.xml",
                f"--root={cwd}",
            ]
        )

    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv))



"""
Tests of large and/or expensive graphs.
"""

import os
import sys
from pathlib import Path

sys.path.append(os.path.join(os.path.dirname(__file__), "../../../tests"))
from gvtest import dot, run_raw  # pylint: disable=wrong-import-position


def test_long_chain():
    """
    This test will fail on 32bit Windows machine if compiled with stack size < 16MB.
    long_chain input file generated using code below:
      from graphviz import Digraph

      graph = Digraph(format="svg")

      prev = "start"
      graph.node("start", label="start")

      for i in range(33000):
        new_node = str(i)
        graph.node(new_node, label=new_node, shape="rectangle")

        graph.edge(prev, new_node)

        prev = new_node

      graph.render("long_chain")
    """
    run_raw("dot", "-Tsvg", "-o", os.devnull, Path(__file__).parent / "long_chain")


def test_wide_clusters():
    """
    A simple regression test for https://gitlab.com/graphviz/graphviz/-/issues/2080#
    """
    dot("svg", Path(__file__).parent / "wide_clusters")



#!/usr/bin/python3

"""
Generate a table of the Graphviz compile configuration for different platforms
from files generated by the ./configure command.
"""

import os
import re
import sys
from argparse import ArgumentParser


def main():
    """entry point"""

    parser = ArgumentParser(
        description="Generate a table of the Graphviz compile "
        "configuration for different platforms from files "
        "generated by the ./configure command."
    )
    parser.add_argument(
        "--color",
        "-c",
        action="store_const",
        dest="colors",
        const="green:red",
        help="Color output using default coloring. Yes is "
        "colored green and No is colored red",
    )
    parser.add_argument(
        "--colors",
        help="Color output using specified COLORS. The format is "
        "<Yes-color>:<No-color>",
    )
    parser.add_argument(
        "--short", "-s", action="store_true", help="Output only Yes or No"
    )
    parser.add_argument("filename", nargs="*", help="Configuration log to read")

    opts = parser.parse_args()

    if opts.colors is None:
        styles = {
            "Yes": "",
            "No": "",
        }
    else:
        yes_color = "green"
        no_color = "red"
        if opts.colors != "":
            colors = opts.colors.split(":")
            if len(colors) == 2:
                yes_color, no_color = colors
            else:
                sys.stderr.write(
                    f"Error: {opts.colors} color specification is illegal\n"
                )
                parser.print_help(file=sys.stderr)
                sys.exit(1)
        styles = {
            "Yes": f' style="color: {yes_color};"',
            "No": f' style="color: {no_color};"',
        }

    table: dict[str, dict[str, dict[str, str]]] = {}
    table_sections: list[str] = []
    component_names: dict[str, list[str]] = {}
    platforms: set[str] = set()

    for filename in opts.filename:
        os_path = os.path.dirname(filename)
        os_version_id = os.path.basename(os_path)
        os_id = os.path.basename(os.path.dirname(os_path))
        platform = f"{os_id.capitalize()} {os_version_id}"
        platforms.add(platform)
        with open(filename, "rt", encoding="utf-8") as fp:
            for line in fp:
                item = [item.strip() for item in line.split(":")]
                if len(item) >= 2:
                    if item[1] == "":
                        section_name = item[0]
                        if section_name not in table:
                            table[section_name] = {}
                            table_sections.append(section_name)
                            component_names[section_name] = []
                    else:
                        component_name, component_value, *_ = item
                        short_value = re.sub("(Yes|No).*", "\\1", component_value)
                        if opts.short:
                            component_value = short_value
                        if component_name not in table[section_name]:
                            table[section_name][component_name] = {}
                            component_names[section_name].append(component_name)
                        table[section_name][component_name][platform] = component_value

    colspan = len(platforms) + 1
    indent = ""
    print(f"{indent}<!DOCTYPE html>")
    print(f"{indent}<html>")
    indent += "  "
    print(f"{indent}<head>")
    indent += "  "
    print(f'{indent}<meta charset="utf-8">')
    print(f"{indent}<style>")
    indent += "  "
    print(
        f"{indent}table {{text-align: left; white-space: nowrap; position: relative;}}"
    )
    print(
        f"{indent}thead tr th {{padding-right: 1em; padding-bottom: 5px; "
        "position: sticky; top: 0px;  background-color: white;}"
    )
    print(f"{indent}td, th {{padding-left: 1.5em;}}")
    indent = indent[:-2]
    print(f"{indent}</style>")
    indent = indent[:-2]
    print(f"{indent}</head>")
    print(f"{indent}<body>")
    indent += "  "
    header = table_sections[0].replace("will be", "was")
    print(f"{indent}<h1>{header}:</h1>")
    print(f"{indent}<table>")
    indent += "  "
    print(f"{indent}<thead>")
    indent += "  "
    print(f"{indent}<tr>")
    indent += "  "
    print(f"{indent}<th></th>")
    for platform in sorted(platforms):
        print(f"{indent}<th>{platform}</th>")
    indent = indent[:-2]
    print(f"{indent} </tr>")
    print(f"{indent} </thead>")
    print(f"{indent} <tbody>")
    indent += "  "
    indent = indent[:-2]
    for section_name in table_sections[1:]:
        print(f'{indent}<tr><th colspan="{colspan}">{section_name};</th></tr>')
        for component_name in component_names[section_name]:
            print(f"{indent}<tr>")
            indent += "  "
            print(f"{indent}<td>{component_name}</td>")
            for platform in sorted(platforms):
                component_value = table[section_name][component_name][platform]
                short_value = re.sub("(Yes|No).*", "\\1", component_value)
                color_style = styles[short_value]
                print(f"{indent}<td{color_style}>{component_value}</td>")
            indent = indent[:-2]
            print(f"{indent}</tr>")
        print(f'{indent}<tr><th colspan="{colspan}">&nbsp;</th></tr>')
    indent = indent[:-2]
    print(f"{indent}</tbody>")
    indent = indent[:-2]
    print(f"{indent}</table>")
    indent = indent[:-2]
    print(f"{indent}</body>")
    indent = indent[:-2]
    print(f"{indent}</html>")


# Python trick to get a main routine
if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""Graphviz CI script for compilation on Windows"""

import argparse
import io
import itertools
import os
import shlex
import shutil
import subprocess
import sys
import textwrap
from pathlib import Path
from typing import Optional, TextIO, Union


def run(
    args: list[Union[str, Path]],
    cwd: Path,
    env: dict[str, str],
    out: Optional[TextIO] = None,
) -> None:
    """run a command, echoing it beforehand"""

    print(f"+ {shlex.join(str(x) for x in args)}", flush=True)
    kwargs = {}
    if out is not None:
        kwargs["stdout"] = subprocess.PIPE
        kwargs["stderr"] = subprocess.STDOUT
    p = subprocess.run(args, cwd=cwd, check=False, text=True, env=env, **kwargs)
    if out is not None:
        sys.stderr.write(p.stdout)
        out.write(p.stdout)
    p.check_returncode()


def require(program: str, fallback: Path, env: dict[str, str], log: TextIO) -> None:
    """
    detect if a given program is missing from the default $Path and needs a fallback

    Args:
        program: The name of a program to look for, without its “.exe” suffix.
        fallback: A path at which to find a version of that program. This alternate
            version will be used if the sought program is not found in $Path.
        env: An environment to use when searching for the program. If the program needs
            a fallback, the $Path variable in this environment will be adjusted.
        log: Sink to write informational messages to.
    """
    assert (
        fallback / f"{program}.exe"
    ).exists(), "{fallback} is not a valid fallback path for {program}"
    which = shutil.which(f"{program}.exe", path=env["PATH"])
    if which is None:
        log.write(
            textwrap.dedent(
                f"""\
        {program} not found
        Fallback needed for: {program}
        Setting up fallback path for: {program} to {fallback}
        """
            )
        )
        env["PATH"] = f"{fallback}{os.pathsep}{env['PATH']}"
    else:
        log.write(f"Found {program} at {which}\n")


def main(args: list[str]) -> int:
    """entry point"""

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "--build-shared-libs",
        choices=("ON", "OFF"),
        default="ON",
        help="control shared libraries selection",
    )
    parser.add_argument(
        "--configuration",
        choices=("Debug", "Release"),
        required=True,
        help="build configuration to select",
    )
    parser.add_argument(
        "--platform", choices=("Win32", "x64"), required=True, help="target platform"
    )
    options = parser.parse_args(args[1:])

    # find the repository root directory
    root = Path(__file__).resolve().parent.parent

    # install Python dependencies
    run([sys.executable, "-m", "pip", "install", "uv"], root, None)
    run(
        [
            sys.executable,
            "-m",
            "uv",
            "pip",
            "install",
            "--requirement",
            "requirements.txt",
        ],
        root,
        None,
    )

    # retrieve submodules, dependencies are stored there
    run(["git", "submodule", "update", "--init", "--depth=1"], root, None)

    # an environment we will use during configuration/compilation
    build_env = os.environ.copy()

    # buffer for output so we can report warning count later
    log = io.StringIO()

    # find some external build dependencies
    utilities = root / "windows/dependencies/graphviz-build-utilities"
    require("win_bison", utilities / "winflexbison", build_env, log)
    require("win_flex", utilities / "winflexbison", build_env, log)
    require("makensis", utilities / "NSIS/Bin", build_env, log)

    build = root / "build"
    if build.exists():
        shutil.rmtree(build)
    build.mkdir(parents=True)
    run(["cmake", "--version"], build, None, log)
    run(
        [
            "cmake",
            "--log-level=VERBOSE",
            "-G",
            "Visual Studio 17 2022",
            "-A",
            options.platform,
            f"-DBUILD_SHARED_LIBS={options.build_shared_libs}",
            "-Dwith_cxx_api=ON",
            "-DENABLE_LTDL=ON",
            "-DWITH_EXPAT=ON",
            "-DWITH_GVEDIT=OFF",
            "-DWITH_ZLIB=ON",
            "--warn-uninitialized",
            "-Werror=dev",
            "..",
        ],
        build,
        build_env,
        log,
    )
    run(
        ["cmake", "--build", ".", "--config", options.configuration],
        build,
        build_env,
        log,
    )
    run(["cpack", "-C", options.configuration], build, build_env, log)

    # report warning count
    warning_count = log.getvalue().count(" warning ")
    metrics = Path("metrics.txt")
    summary = f"{os.environ['CI_JOB_NAME']}-warnings {warning_count}"
    print(summary, flush=True)
    metrics.write_text(f"{summary}\n", encoding="utf-8")

    # derive Graphviz version
    version_buffer = io.StringIO()
    run([sys.executable, "gen_version.py"], root, None, version_buffer)
    gv_version = version_buffer.getvalue().strip()

    # find the installer
    installers = list(build.glob("Graphviz*.exe"))
    assert len(installers) == 1, "failed to find Graphviz installer"
    installer = build / installers[0]

    # install Graphviz
    install_dir = "C:\\Graphviz"
    run([installer, "/S", f"/D={install_dir}"], root, None)

    # which Windows interface are we targeting?
    if options.platform == "x64":
        api = "win64"
    else:
        api = "win32"

    # move the installer to the location expected by CI archiving steps
    dst = build / f"graphviz-install-{gv_version}-{api}.exe"
    shutil.move(installer, dst)

    # an environment we will use during testing
    test_env = os.environ.copy()

    test_env["PATH"] = f"{install_dir}\\bin;{test_env.get('PATH', '')}"
    test_env["CFLAGS"] = f"{test_env.get('CFLAGS', '')} -I{install_dir}\\include"
    test_env["LIB"] = f"{test_env.get('LIB', '')};{install_dir}\\lib"
    test_env["graphviz_ROOT"] = install_dir

    # run the test suite
    run(
        [
            sys.executable,
            "-m",
            "pytest",
            "-m",
            "not slow",
            "-n",
            "auto",
            "--junit-xml=report.xml",
            "ci/tests.py",
            "tests",
        ],
        root,
        test_env,
    )

    # create artifacts to archive
    prefix = f"windows_10_cmake_{os.environ['configuration']}_"
    packages = root / "Packages" / os.environ["CI_JOB_NAME"]
    packages.mkdir(parents=True)
    for src in itertools.chain(build.glob("*.exe"), build.glob("*.zip")):
        dst = packages / f"{prefix}{src.name}"
        print(f"+ mv {shlex.join(str(a) for a in [src, dst])}", flush=True)
        shutil.move(src, dst)

    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv))



#!/usr/bin/env python3

"""
(brewer_colors, svgcolor_names, color_names) → colortbl.h generator
"""

import argparse
import locale
import re
import sys


def rgb_to_hsv(r: int, g: int, b: int) -> tuple[int, int, int]:
    """transform an RGB color into HSV format"""

    r /= 255.0
    g /= 255.0
    b /= 255.0

    max_ = max((r, g, b))
    min_ = min((r, g, b))

    v = max_

    if max_ != 0:
        s = (max_ - min_) / max_
    else:
        s = 0

    if s == 0:
        h = 0
    else:
        delta = max_ - min_
        rc = (max_ - r) / delta
        gc = (max_ - g) / delta
        bc = (max_ - b) / delta
        if r == max_:
            h = bc - gc
        elif g == max_:
            h = 2.0 + (rc - bc)
        else:
            h = 4.0 + (gc - rc)
        h *= 60.0
        if h < 0.0:
            h += 360.0

    return int(h / 360.0 * 255.0), int(s * 255.0), int(v * 255.0)


def main(args: list[str]) -> int:
    """entry point"""

    # avoid sorting issues due to locale differences
    locale.setlocale(locale.LC_ALL, "C")

    # parse command line arguments
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "brewer_colors",
        type=argparse.FileType("rt"),
        help="input Brewer color table entries",
    )
    parser.add_argument(
        "svgcolor_names",
        type=argparse.FileType("rt"),
        help="input SVG color table entries",
    )
    parser.add_argument(
        "color_names",
        type=argparse.FileType("rt"),
        help="input other color table entries",
    )
    parser.add_argument(
        "output", type=argparse.FileType("wt"), help="output color table C header"
    )
    options = parser.parse_args(args[1:])

    options.output.write("static const hsvrgbacolor_t color_lib[] = {\n")

    # collect all color entries
    entries = (
        options.brewer_colors.readlines()
        + options.svgcolor_names.readlines()
        + options.color_names.readlines()
    )

    # sort these so `bsearch` for a color will work correctly
    entries = sorted(entries, key=locale.strxfrm)

    for entry in entries:
        # ignore blank lines
        if entry.strip() == "":
            continue
        # ignore commented lines
        if re.match(r"\s*#", entry) is not None:
            continue

        # extract fields of the entry
        m = re.match(
            r"(?P<name>\S+) (?P<r>\d+) (?P<g>\d+) (?P<b>\d+) (?P<a>\d+)", entry
        )
        assert m is not None, f'unexpected entry format "{entry}"'

        # transform an RGB color into HSV
        r = int(m.group("r"))
        g = int(m.group("g"))
        b = int(m.group("b"))
        h, s, v = rgb_to_hsv(r, g, b)

        # write this out as a C array entry
        name = m.group("name")
        a = int(m.group("a"))
        options.output.write(f'{{"{name}",{h},{s},{v},{r},{g},{b},{a}}},\n')

    options.output.write("};\n")

    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv))



"""
Tests of Graphviz output against expected reference output.
"""

import difflib
import os
import sys
from pathlib import Path

import pytest

sys.path.append(os.path.join(os.path.dirname(__file__), "../../tests"))
from gvtest import run, run_raw  # pylint: disable=wrong-import-position


@pytest.mark.xfail
@pytest.mark.parametrize(
    "src,format,reference",
    (
        ('graph G {a [label="" shape=none]}', "gv", "001.001.gv"),
        ('graph G {a [label="" shape=none]}', "xdot", "001.002.xdot"),
        ('graph G {a [label="" shape=none]}', "png", "001.003.png"),
        ('graph G {a [label="" shape=none]}', "svg", "001.004.svg"),
        ('graph G {a [label="" shape=none]}', "ps", "001.005.ps"),
        ('graph G {a [label="" shape=none]}', "cmapx", "001.006.cmapx"),
        ('graph G {a [label="" shape=plaintext]}', "xdot", "002.001.xdot"),
        ('graph G {a [label="" shape=point]}', "xdot", "003.001.xdot"),
        ('graph G {a [label="" shape=box]}', "xdot", "004.001.xdot"),
        ('graph G {a [label="" shape=rect]}', "xdot", "005.001.xdot"),
        ('graph G {a [label="" shape=rectangle]}', "xdot", "006.001.xdot"),
        ('graph G {a [label="" shape=ellipse]}', "xdot", "007.001.xdot"),
        ('graph G {a [label="" shape=circle]}', "xdot", "008.001.xdot"),
        ('graph G {a [label="" shape=triangle]}', "xdot", "009.001.xdot"),
        ('graph G {a [label="" shape=invtriangle]}', "xdot", "010.001.xdot"),
        ('graph G {a [label="" shape=diamond]}', "xdot", "011.001.xdot"),
        ('graph G {a [label="" shape=trapezium]}', "xdot", "012.001.xdot"),
        ('graph G {a [label="" shape=invtrapezium]}', "xdot", "013.001.xdot"),
        ('graph G {a [label="" shape=parallelogram]}', "xdot", "014.001.xdot"),
        ('graph G {a [label="" shape=house]}', "xdot", "015.001.xdot"),
        ('graph G {a [label="" shape=invhouse]}', "xdot", "016.001.xdot"),
        ('graph G {a [label="" shape=pentagon]}', "xdot", "017.001.xdot"),
        ('graph G {a [label="" shape=hexagon]}', "xdot", "018.001.xdot"),
        ('graph G {a [label="" shape=septagon]}', "xdot", "019.001.xdot"),
        ('graph G {a [label="" shape=octagon]}', "xdot", "020.001.xdot"),
        ('graph G {a [label="" shape=doubleoctagon]}', "xdot", "021.001.xdot"),
        ('graph G {a [label="" shape=doublecircle]}', "xdot", "022.001.xdot"),
        ('graph G {a [label="" shape=tripleoctagon]}', "xdot", "023.001.xdot"),
        ('graph G {a [label="" shape=Mdiamond]}', "xdot", "024.001.xdot"),
        ('graph G {a [label="" shape=Msquare]}', "xdot", "025.001.xdot"),
        ('graph G {a [label="" shape=Mcircle]}', "xdot", "026.001.xdot"),
        ('graph G {a [label="" shape=none]}', "xdot", "027.001.xdot"),
        ("graph G {a [label=b shape=none]}", "xdot", "028.001.xdot"),
        ("graph G {a [shape=none]}", "xdot", "029.001.xdot"),
        ('graph G {a [label="\\N" shape=none]}', "xdot", "030.001.xdot"),
        ('graph G {a [label="n_of_\\G" shape=none]}', "xdot", "031.001.xdot"),
        ('graph G {a [label="\\E\\H\\T" shape=none]}', "xdot", "032.001.xdot"),
        (
            'digraph G {node [label="" shape=none]; a->b [label="\\E"]}',
            "xdot",
            "033.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [label="\\T"]}',
            "xdot",
            "034.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [label="\\H"]}',
            "xdot",
            "035.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [label="e_of_\\G"]}',
            "xdot",
            "036.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [label="\\N"]}',
            "xdot",
            "037.001.xdot",
        ),
        (
            'graph G {a [label="center\\naligned\\ntext" shape=none]}',
            "xdot",
            "038.001.xdot",
        ),
        (
            'graph G {a [label="left\\laligned\\ltext" shape=none]}',
            "xdot",
            "039.001.xdot",
        ),
        (
            'graph G {a [label="right\\raligned\\rtext" shape=none]}',
            "xdot",
            "040.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=none]}',
            "xdot",
            "041.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=""]}',
            "xdot",
            "042.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=normal]}',
            "xdot",
            "043.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=box]}',
            "xdot",
            "044.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=crow]}',
            "xdot",
            "045.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=diamond]}',
            "xdot",
            "046.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=dot]}',
            "xdot",
            "047.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=inv]}',
            "xdot",
            "048.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=tee]}',
            "xdot",
            "049.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=vee]}',
            "xdot",
            "050.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=lnormal]}',
            "xdot",
            "051.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=lbox]}',
            "xdot",
            "052.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=lcrow]}',
            "xdot",
            "053.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=ldiamond]}',
            "xdot",
            "054.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=ldot]}',
            "xdot",
            "055.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=linv]}',
            "xdot",
            "056.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=ltee]}',
            "xdot",
            "057.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=lvee]}',
            "xdot",
            "058.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=rnormal]}',
            "xdot",
            "059.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=rbox]}',
            "xdot",
            "060.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=rcrow]}',
            "xdot",
            "061.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=rdiamond]}',
            "xdot",
            "062.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=rdot]}',
            "xdot",
            "063.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=rinv]}',
            "xdot",
            "064.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=rtee]}',
            "xdot",
            "065.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=rvee]}',
            "xdot",
            "066.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=onormal]}',
            "xdot",
            "067.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=obox]}',
            "xdot",
            "068.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=ocrow]}',
            "xdot",
            "069.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=odiamond]}',
            "xdot",
            "070.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=odot]}',
            "xdot",
            "071.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=oinv]}',
            "xdot",
            "072.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=otee]}',
            "xdot",
            "073.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=ovee]}',
            "xdot",
            "074.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=onormalnormal]}',
            "xdot",
            "075.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=oboxbox]}',
            "xdot",
            "076.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=ocrowcrow]}',
            "xdot",
            "077.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=odiamonddiamond]}',
            "xdot",
            "078.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=odotdot]}',
            "xdot",
            "079.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=oinvinv]}',
            "xdot",
            "080.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=oteetee]}',
            "xdot",
            "081.001.xdot",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [arrowhead=oveevee]}',
            "xdot",
            "082.001.xdot",
        ),
        (
            'graph G {a [label="" shape=box style=filled color=black]}',
            "xdot",
            "083.001.xdot",
        ),
        (
            'graph G {a [label="" shape=box style=filled color=black]}',
            "png",
            "083.002.png",
        ),
        (
            'graph G {a [label="" shape=box style=filled color=red]}',
            "xdot",
            "084.001.xdot",
        ),
        (
            'graph G {a [label="" shape=box style=filled color=red]}',
            "png",
            "084.002.png",
        ),
        (
            'graph G {a [label="" shape=box style=filled color=hotpink]}',
            "xdot",
            "085.001.xdot",
        ),
        (
            'graph G {a [label="" shape=box style=filled color=hotpink]}',
            "png",
            "085.002.png",
        ),
        (
            'graph G {a [label="" shape=box style=filled color=transparent]}',
            "xdot",
            "086.001.xdot",
        ),
        (
            'graph G {a [label="" shape=box style=filled color=transparent]}',
            "png",
            "086.002.png",
        ),
        ('graph G {a [URL="" label="" shape=none]}', "cmapx", "087.001.cmapx"),
        ('graph G {a [URL="" label="" shape=none]}', "ps", "087.002.ps"),
        ('graph G {a [URL="" label="" shape=none]}', "svg", "087.003.svg"),
        ('graph G {graph [URL="\\G.html"]}', "cmapx", "088.001.cmapx"),
        ('graph G {graph [URL="\\G.html"]}', "ps", "088.002.ps"),
        ('graph G {graph [URL="\\G.html"]}', "svg", "088.003.svg"),
        ('graph G {a [URL="\\N.html" label="" shape=none]}', "cmapx", "089.001.cmapx"),
        ('graph G {a [URL="\\N.html" label="" shape=none]}', "ps", "089.002.ps"),
        ('graph G {a [URL="\\N.html" label="" shape=none]}', "svg", "089.003.svg"),
        (
            'digraph G {node [label="" shape=none]; a->b [URL="\\E.html"]}',
            "cmapx",
            "090.001.cmapx",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [URL="\\E.html"]}',
            "ps",
            "090.002.ps",
        ),
        (
            'digraph G {node [label="" shape=none]; a->b [URL="\\E.html"]}',
            "svg",
            "090.003.svg",
        ),
    ),
)
def test_reference(src: str, format: str, reference: str, tmp_path: Path):
    """test some Graphviz input against a reference source"""

    # locate the reference output
    ref = Path(__file__).resolve().parent / "test_reference" / reference
    assert ref.exists()

    # process our input with Graphviz
    output = tmp_path / reference
    run("dot", f"-T{format}", "-o", output, input=src)

    # compare against the reference output
    if output.suffix == ".png":
        run_raw("diff", ref, output)
    else:
        fail = False
        a = ref.read_text(encoding="utf-8")
        b = output.read_text(encoding="utf-8")
        for line in difflib.unified_diff(a, b, fromfile=str(ref), tofile=str(output)):
            sys.stderr.write(f"{line}\n")
            fail = True
        if fail:
            raise RuntimeError(f"diff {ref} {output} failed")
